@@ -1,5 +1,6 @@
 /*
  * Copyright (C) Mellanox Technologies Ltd. 2001-2017. ALL RIGHTS RESERVED.
+ * Copyright (c) 2022      IBM Corporation.  All rights reserved.
  * $COPYRIGHT$
  *
  * Additional copyrights may follow
@@ -17,13 +18,26 @@
 #include "osc_ucx.h"
 #include "osc_ucx_request.h"
 
-#define UCX_VERSION(_major, _minor, _build) (((_major) * 100) + (_minor))
-
 #define memcpy_off(_dst, _src, _len, _off)        \
     memcpy(((char*)(_dst)) + (_off), _src, _len); \
     (_off) += (_len);
 
+opal_mutex_t mca_osc_service_mutex = OPAL_MUTEX_STATIC_INIT;
+static void _osc_ucx_init_lock(void)
+{
+    if(mca_osc_ucx_component.enable_mpi_threads) {
+        opal_mutex_lock(&mca_osc_service_mutex);
+    }
+}
+static void _osc_ucx_init_unlock(void)
+{
+    if(mca_osc_ucx_component.enable_mpi_threads) {
+        opal_mutex_unlock(&mca_osc_service_mutex);
+    }
+}
+
 static int component_open(void);
+static int component_close(void);
 static int component_register(void);
 static int component_init(bool enable_progress_threads, bool enable_mpi_threads);
 static int component_finalize(void);
@@ -42,6 +56,7 @@ ompi_osc_ucx_component_t mca_osc_ucx_component = {
             MCA_BASE_MAKE_VERSION(component, OMPI_MAJOR_VERSION, OMPI_MINOR_VERSION,
                                   OMPI_RELEASE_VERSION),
             .mca_open_component = component_open,
+            .mca_close_component = component_close,
             .mca_register_component_params = component_register,
         },
         .osc_data = {
@@ -100,6 +115,14 @@ ompi_osc_ucx_module_t ompi_osc_ucx_module_template = {
 };
 
 static int component_open(void) {
+    opal_common_ucx_mca_register();
+
+    return OMPI_SUCCESS;
+}
+
+static int component_close(void) {
+    opal_common_ucx_mca_deregister();
+
     return OMPI_SUCCESS;
 }
 
@@ -130,11 +153,76 @@ static int progress_callback(void) {
     return 0;
 }
 
+static int ucp_context_init(void) {
+    int ret = OMPI_SUCCESS;
+    ucs_status_t status;
+    ucp_config_t *config = NULL;
+    ucp_params_t context_params;
+
+    status = ucp_config_read("MPI", NULL, &config);
+    if (UCS_OK != status) {
+        OSC_UCX_VERBOSE(1, "ucp_config_read failed: %d", status);
+        return OMPI_ERROR;
+    }
+
+    /* initialize UCP context */
+    memset(&context_params, 0, sizeof(context_params));
+    context_params.field_mask = UCP_PARAM_FIELD_FEATURES |
+                                UCP_PARAM_FIELD_MT_WORKERS_SHARED |
+                                UCP_PARAM_FIELD_ESTIMATED_NUM_EPS |
+                                UCP_PARAM_FIELD_REQUEST_INIT |
+                                UCP_PARAM_FIELD_REQUEST_SIZE;
+    context_params.features = UCP_FEATURE_RMA | UCP_FEATURE_AMO32 | UCP_FEATURE_AMO64;
+    context_params.mt_workers_shared = 0;
+    context_params.estimated_num_eps = ompi_proc_world_size();
+    context_params.request_init = internal_req_init;
+    context_params.request_size = sizeof(ompi_osc_ucx_internal_request_t);
+
+    status = ucp_init(&context_params, config, &mca_osc_ucx_component.ucp_context);
+    ucp_config_release(config);
+    if (UCS_OK != status) {
+        OSC_UCX_VERBOSE(1, "ucp_init failed: %d", status);
+        ret =  OMPI_ERROR;
+    }
+
+    return ret;
+}
 static int component_init(bool enable_progress_threads, bool enable_mpi_threads) {
+    opal_common_ucx_support_level_t support_level = OPAL_COMMON_UCX_SUPPORT_NONE;
+    mca_base_var_source_t param_source = MCA_BASE_VAR_SOURCE_DEFAULT;
+    int ret = OMPI_SUCCESS,
+        param = -1;
+
     mca_osc_ucx_component.enable_mpi_threads = enable_mpi_threads;
 
-    opal_common_ucx_mca_register();
-    return OMPI_SUCCESS;
+    ret = ucp_context_init();
+    if (OMPI_ERROR == ret) {
+        return OMPI_ERR_NOT_AVAILABLE;
+    }
+
+    support_level = opal_common_ucx_support_level(mca_osc_ucx_component.ucp_context);
+    if (OPAL_COMMON_UCX_SUPPORT_NONE == support_level) {
+        ucp_cleanup(mca_osc_ucx_component.ucp_context);
+        mca_osc_ucx_component.ucp_context = NULL;
+        return OMPI_ERR_NOT_AVAILABLE;
+    }
+
+    param = mca_base_var_find("ompi","osc","ucx","priority");
+    if (0 <= param) {
+        (void) mca_base_var_get_value(param, NULL, &param_source, NULL);
+    }
+
+    /*
+     * Retain priority if we have supported devices and transports.
+     * Lower priority if we have supported transports, but not supported devices.
+     */
+    if (MCA_BASE_VAR_SOURCE_DEFAULT == param_source) {
+        mca_osc_ucx_component.priority = (support_level == OPAL_COMMON_UCX_SUPPORT_DEVICE) ?
+                    mca_osc_ucx_component.priority : 9;
+        OSC_UCX_VERBOSE(2, "returning priority %d", mca_osc_ucx_component.priority);
+    }
+
+    return ret;
 }
 
 static int component_finalize(void) {
@@ -148,15 +236,20 @@ static int component_finalize(void) {
 
     if (mca_osc_ucx_component.ucp_worker != NULL) {
         ucp_worker_destroy(mca_osc_ucx_component.ucp_worker);
+        mca_osc_ucx_component.ucp_worker = NULL;
     }
 
     assert(mca_osc_ucx_component.num_incomplete_req_ops == 0);
     if (mca_osc_ucx_component.env_initialized == true) {
         OBJ_DESTRUCT(&mca_osc_ucx_component.requests);
-        ucp_cleanup(mca_osc_ucx_component.ucp_context);
         mca_osc_ucx_component.env_initialized = false;
     }
-    opal_common_ucx_mca_deregister();
+
+    if (NULL != mca_osc_ucx_component.ucp_context) {
+        ucp_cleanup(mca_osc_ucx_component.ucp_context);
+        mca_osc_ucx_component.ucp_context = NULL;
+    }
+
     return OMPI_SUCCESS;
 }
 
@@ -256,6 +349,9 @@ static void ompi_osc_ucx_unregister_progress()
 {
     int ret;
 
+    /* May be called concurrently - protect */
+    _osc_ucx_init_lock();
+
     mca_osc_ucx_component.num_modules--;
     OSC_UCX_ASSERT(mca_osc_ucx_component.num_modules >= 0);
     if (0 == mca_osc_ucx_component.num_modules) {
@@ -264,6 +360,8 @@ static void ompi_osc_ucx_unregister_progress()
             OSC_UCX_VERBOSE(1, "opal_progress_unregister failed: %d", ret);
         }
     }
+
+    _osc_ucx_init_unlock();
 }
 
 static int component_select(struct ompi_win_t *win, void **base, size_t size, int disp_unit,
@@ -297,18 +395,12 @@ static int component_select(struct ompi_win_t *win, void **base, size_t size, in
         return OMPI_ERR_NOT_SUPPORTED;
     }
 
+    _osc_ucx_init_lock();
+
     if (mca_osc_ucx_component.env_initialized == false) {
-        ucp_config_t *config = NULL;
-        ucp_params_t context_params;
         ucp_worker_params_t worker_params;
         ucp_worker_attr_t worker_attr;
 
-        status = ucp_config_read("MPI", NULL, &config);
-        if (UCS_OK != status) {
-            OSC_UCX_VERBOSE(1, "ucp_config_read failed: %d", status);
-            return OMPI_ERROR;
-        }
-
         OBJ_CONSTRUCT(&mca_osc_ucx_component.requests, opal_free_list_t);
         ret = opal_free_list_init (&mca_osc_ucx_component.requests,
                                    sizeof(ompi_osc_ucx_request_t),
@@ -317,42 +409,24 @@ static int component_select(struct ompi_win_t *win, void **base, size_t size, in
                                    0, 0, 8, 0, 8, NULL, 0, NULL, NULL, NULL);
         if (OMPI_SUCCESS != ret) {
             OSC_UCX_VERBOSE(1, "opal_free_list_init failed: %d", ret);
-            goto error;
-        }
-
-        /* initialize UCP context */
-
-        memset(&context_params, 0, sizeof(context_params));
-        context_params.field_mask = UCP_PARAM_FIELD_FEATURES |
-                                    UCP_PARAM_FIELD_MT_WORKERS_SHARED |
-                                    UCP_PARAM_FIELD_ESTIMATED_NUM_EPS |
-                                    UCP_PARAM_FIELD_REQUEST_INIT |
-                                    UCP_PARAM_FIELD_REQUEST_SIZE;
-        context_params.features = UCP_FEATURE_RMA | UCP_FEATURE_AMO32 | UCP_FEATURE_AMO64;
-        context_params.mt_workers_shared = 0;
-        context_params.estimated_num_eps = ompi_proc_world_size();
-        context_params.request_init = internal_req_init;
-        context_params.request_size = sizeof(ompi_osc_ucx_internal_request_t);
-
-        status = ucp_init(&context_params, config, &mca_osc_ucx_component.ucp_context);
-        ucp_config_release(config);
-        if (UCS_OK != status) {
-            OSC_UCX_VERBOSE(1, "ucp_init failed: %d", status);
-            ret = OMPI_ERROR;
-            goto error;
+            goto select_unlock;
         }
 
         assert(mca_osc_ucx_component.ucp_worker == NULL);
         memset(&worker_params, 0, sizeof(worker_params));
         worker_params.field_mask = UCP_WORKER_PARAM_FIELD_THREAD_MODE;
-        worker_params.thread_mode = (mca_osc_ucx_component.enable_mpi_threads == true)
-                                    ? UCS_THREAD_MODE_MULTI : UCS_THREAD_MODE_SINGLE;
+        if (mca_osc_ucx_component.enable_mpi_threads) {
+            worker_params.thread_mode = UCS_THREAD_MODE_MULTI;
+        } else {
+            worker_params.thread_mode =
+                opal_common_ucx_thread_mode(ompi_mpi_thread_provided);
+        }
         status = ucp_worker_create(mca_osc_ucx_component.ucp_context, &worker_params,
                                    &(mca_osc_ucx_component.ucp_worker));
         if (UCS_OK != status) {
             OSC_UCX_VERBOSE(1, "ucp_worker_create failed: %d", status);
-            ret = OMPI_ERROR;
-            goto error_nomem;
+            ret = OMPI_ERR_TEMP_OUT_OF_RESOURCE;
+            goto select_unlock;
         }
 
         /* query UCP worker attributes */
@@ -360,20 +434,44 @@ static int component_select(struct ompi_win_t *win, void **base, size_t size, in
         status = ucp_worker_query(mca_osc_ucx_component.ucp_worker, &worker_attr);
         if (UCS_OK != status) {
             OSC_UCX_VERBOSE(1, "ucp_worker_query failed: %d", status);
-            ret = OMPI_ERROR;
-            goto error_nomem;
+            ret = OMPI_ERR_TEMP_OUT_OF_RESOURCE;
+            goto select_unlock;
         }
 
         if (mca_osc_ucx_component.enable_mpi_threads == true &&
             worker_attr.thread_mode != UCS_THREAD_MODE_MULTI) {
             OSC_UCX_VERBOSE(1, "ucx does not support multithreading");
-            ret = OMPI_ERROR;
-            goto error_nomem;
+            ret = OMPI_ERR_TEMP_OUT_OF_RESOURCE;
+            goto select_unlock;
         }
 
         mca_osc_ucx_component.env_initialized = true;
         env_initialized = true;
     }
+    
+    mca_osc_ucx_component.num_modules++;
+
+    OSC_UCX_ASSERT(mca_osc_ucx_component.num_modules > 0);
+    if (1 == mca_osc_ucx_component.num_modules) {
+        ret = opal_progress_register(progress_callback);
+        if (OMPI_SUCCESS != ret) {
+            OSC_UCX_VERBOSE(1, "opal_progress_register failed: %d", ret);
+            goto select_unlock;
+        }
+    }
+
+select_unlock:
+    _osc_ucx_init_unlock();
+    switch(ret) {
+    case OMPI_SUCCESS:
+        break;
+    case OMPI_ERROR:
+        goto error;
+    case OMPI_ERR_TEMP_OUT_OF_RESOURCE:
+        goto error_nomem;
+    default:
+        goto error;
+    }
 
     /* create module structure */
     module = (ompi_osc_ucx_module_t *)calloc(1, sizeof(ompi_osc_ucx_module_t));
@@ -382,7 +480,6 @@ static int component_select(struct ompi_win_t *win, void **base, size_t size, in
         goto error_nomem;
     }
 
-    mca_osc_ucx_component.num_modules++;
 
     /* fill in the function pointer part */
     memcpy(module, &ompi_osc_ucx_module_template, sizeof(ompi_osc_base_module_t));
@@ -650,14 +747,6 @@ static int component_select(struct ompi_win_t *win, void **base, size_t size, in
         goto error;
     }
 
-    OSC_UCX_ASSERT(mca_osc_ucx_component.num_modules > 0);
-    if (1 == mca_osc_ucx_component.num_modules) {
-        ret = opal_progress_register(progress_callback);
-        if (OMPI_SUCCESS != ret) {
-            OSC_UCX_VERBOSE(1, "opal_progress_register failed: %d", ret);
-            goto error;
-        }
-    }
     return ret;
 
  error:
@@ -693,8 +782,13 @@ static int component_select(struct ompi_win_t *win, void **base, size_t size, in
 error_nomem:
     if (env_initialized == true) {
         OBJ_DESTRUCT(&mca_osc_ucx_component.requests);
+
         ucp_worker_destroy(mca_osc_ucx_component.ucp_worker);
+        mca_osc_ucx_component.ucp_worker = NULL;
+
         ucp_cleanup(mca_osc_ucx_component.ucp_context);
+        mca_osc_ucx_component.ucp_context = NULL;
+
         mca_osc_ucx_component.env_initialized = false;
     }
     return ret;
@@ -819,6 +913,7 @@ int ompi_osc_ucx_win_detach(struct ompi_win_t *win, const void *base) {
 int ompi_osc_ucx_free(struct ompi_win_t *win) {
     ompi_osc_ucx_module_t *module = (ompi_osc_ucx_module_t*) win->w_osc_module;
     int i, ret;
+    uint64_t idx;
 
     assert(module->global_ops_num == 0);
     assert(module->lock_count == 0);
@@ -839,6 +934,14 @@ int ompi_osc_ucx_free(struct ompi_win_t *win) {
     ret = module->comm->c_coll->coll_barrier(module->comm,
                                              module->comm->c_coll->coll_barrier_module);
 
+    /* MPI_Win_free should detach any memory attached to dynamic windows */
+    for (idx = 0; idx < module->state.dynamic_win_count; idx++) {
+        assert(module->local_dynamic_win_info[idx].refcnt == 1);
+        ucp_mem_unmap(mca_osc_ucx_component.ucp_context,
+                      module->local_dynamic_win_info[idx].memh);
+    }
+    module->state.dynamic_win_count = 0;
+
     for (i = 0; i < ompi_comm_size(module->comm); i++) {
         if ((module->win_info_array[i]).rkey_init == true) {
             ucp_rkey_destroy((module->win_info_array[i]).rkey);
