@@ -2,6 +2,11 @@
 /*
  * Copyright (c) 2012-2013 Los Alamos National Security, LLC. All rights
  *                         reserved.
+ * Copyright (c) 2014 Cisco Systems, Inc.  All rights reserved.
+ * Copyright (c) 2017      IBM Corporation. All rights reserved.
+ * Copyright (c) 2020      The University of Tennessee and The University
+ *                         of Tennessee Research Foundation.  All rights
+ *                         reserved.
  * $COPYRIGHT$
  *
  * Additional copyrights may follow
@@ -19,7 +24,6 @@
 #include "ompi/mpi/tool/profile/defines.h"
 #endif
 
-static const char FUNC_NAME[] = "MPI_T_cvar_write";
 
 int MPI_T_cvar_write (MPI_T_cvar_handle handle, const void *buf)
 {
@@ -30,10 +34,10 @@ int MPI_T_cvar_write (MPI_T_cvar_handle handle, const void *buf)
     }
 
     if (MPI_PARAM_CHECK && NULL == buf) {
-        return MPI_ERR_ARG;
+        return MPI_T_ERR_INVALID;
     }
 
-    mpit_lock ();
+    ompi_mpit_lock ();
 
     do {
         if (MCA_BASE_VAR_SCOPE_CONSTANT == handle->var->mbv_scope ||
@@ -53,7 +57,7 @@ int MPI_T_cvar_write (MPI_T_cvar_handle handle, const void *buf)
         }
     } while (0);
 
-    mpit_unlock ();
+    ompi_mpit_unlock ();
 
     return rc;
 }
