@@ -9,8 +9,10 @@
  * Copyright (c) 2011      Oracle and/or all its affiliates.  All rights reserved.
  * Copyright (c) 2011-2013 Los Alamos National Security, LLC.
  *                         All rights reserved.
- * Copyright (c) 2014-2017 Intel, Inc.  All rights reserved.
- * Copyright (c) 2017      IBM Corporation.  All rights reserved.
+ * Copyright (c) 2014-2018 Intel, Inc.  All rights reserved.
+ * Copyright (c) 2017-2022 IBM Corporation.  All rights reserved.
+ * Copyright (c) 2018      Research Organization for Information Science
+ *                         and Technology (RIST).  All rights reserved.
  * $COPYRIGHT$
  *
  * Additional copyrights may follow
@@ -49,7 +51,6 @@
 #include "orte/util/name_fns.h"
 #include "orte/util/proc_info.h"
 #include "orte/util/show_help.h"
-#include "orte/util/nidmap.h"
 #include "orte/util/threads.h"
 
 #include "orte/runtime/orte_globals.h"
@@ -148,14 +149,16 @@ static void hnp_abort(int error_code, char *fmt, ...)
     orte_abnormal_term_ordered = true;
 
     /* If there was a message, construct it */
-    va_start(arglist, fmt);
-    if (NULL != fmt) {
-        vasprintf(&outmsg, fmt, arglist);
-    }
-    va_end(arglist);
+    if (!orte_do_not_launch) {
+        va_start(arglist, fmt);
+        if (NULL != fmt) {
+            vasprintf(&outmsg, fmt, arglist);
+        }
+        va_end(arglist);
 
-    /* use the show-help system to get the message out */
-    orte_show_help("help-errmgr-base.txt", "simple-message", true, outmsg);
+        /* use the show-help system to get the message out */
+        orte_show_help("help-errmgr-base.txt", "simple-message", true, outmsg);
+    }
 
     /* this could have happened very early, so see if it happened
      * before we started anything - if so, we can just finalize */
@@ -840,14 +843,11 @@ static void default_hnp_abort(orte_job_t *jdata)
     i32ptr = &i32;
     if (orte_get_attribute(&jdata->attributes, ORTE_JOB_NUM_NONZERO_EXIT, (void**)&i32ptr, OPAL_INT32)) {
         /* warn user */
-        opal_output(orte_clean_output,
-                    "-------------------------------------------------------\n"
-                    "%s job %s terminated normally, but %d %s. Per user-direction, the job has been aborted.\n"
-                    "-------------------------------------------------------",
-                    (1 == ORTE_LOCAL_JOBID(jdata->jobid)) ? "Primary" : "Child",
-                    (1 == ORTE_LOCAL_JOBID(jdata->jobid)) ? "" : ORTE_LOCAL_JOBID_PRINT(jdata->jobid),
-                    i32, (1 == i32) ? "process returned\na non-zero exit code" :
-                    "processes returned\nnon-zero exit codes");
+        orte_show_help("help-errmgr-base.txt", "normal-termination-but", true, 
+                       (1 == ORTE_LOCAL_JOBID(jdata->jobid)) ? "Primary" : "Child",
+                       (1 == ORTE_LOCAL_JOBID(jdata->jobid)) ? "" : ORTE_LOCAL_JOBID_PRINT(jdata->jobid),
+                       i32, (1 == i32) ? "process returned\na non-zero exit code" :
+                       "processes returned\nnon-zero exit codes");
     }
 
     OPAL_OUTPUT_VERBOSE((1, orte_errmgr_base_framework.framework_output,
