@@ -13,10 +13,11 @@
  * Copyright (c) 2007-2012 Los Alamos National Security, LLC.  All rights
  *                         reserved.
  * Copyright (c) 2008-2009 Sun Microsystems, Inc.  All rights reserved.
- * Copyright (c) 2011-2017 IBM Corporation.  All rights reserved.
- * Copyright (c) 2014-2018 Intel, Inc. All rights reserved.
+ * Copyright (c) 2011-2019 IBM Corporation.  All rights reserved.
+ * Copyright (c) 2014-2020 Intel, Inc.  All rights reserved.
  * Copyright (c) 2015-2018 Research Organization for Information Science
  *                         and Technology (RIST). All rights reserved.
+ * Copyright (c) 2023      Jeffrey M. Squyres.  All rights reserved.
  * $COPYRIGHT$
  *
  * Additional copyrights may follow
@@ -343,11 +344,13 @@ static int setup_launch(int *argcptr, char ***argvptr,
     char *orted_cmd, *orted_prefix, *final_cmd;
     int orted_index;
     int rc;
-    int i, j;
+    int i, j, cnt;
     bool found;
     char *lib_base=NULL, *bin_base=NULL;
     char *opal_prefix = getenv("OPAL_PREFIX");
+    char *opal_libdir = getenv("OPAL_LIBDIR");
     char* full_orted_cmd = NULL;
+    char * rtmod;
 
     /* Figure out the basenames for the libdir and bindir.  This
        requires some explanation:
@@ -473,7 +476,12 @@ static int setup_launch(int *argcptr, char ***argvptr,
          */
 
         value = opal_basename(opal_install_dirs.bindir);
-        asprintf(&bin_base, "%s/%s", prefix_dir, value);
+        /* Ensure that we don't have a prefix that ends in "/" */
+        if ('/' == prefix_dir[strlen(prefix_dir) - 1]) {
+            asprintf(&bin_base, "%s%s", prefix_dir, value);
+        } else {
+            asprintf(&bin_base, "%s/%s", prefix_dir, value);
+        }
         free(value);
 
         if (NULL != orted_cmd) {
@@ -500,13 +508,19 @@ static int setup_launch(int *argcptr, char ***argvptr,
              * we have to insert the orted_prefix in the right place
              */
             (void)asprintf (&final_cmd,
-                            "%s%s%s PATH=%s%s$PATH ; export PATH ; "
-                            "LD_LIBRARY_PATH=%s%s$LD_LIBRARY_PATH ; export LD_LIBRARY_PATH ; "
-                            "DYLD_LIBRARY_PATH=%s%s$DYLD_LIBRARY_PATH ; export DYLD_LIBRARY_PATH ; "
+                            "%s%s%s%s%s%s%s%s%s PATH=%s%s$PATH ; export PATH ; "
+                            "LD_LIBRARY_PATH=%s%s${LD_LIBRARY_PATH:-} ; export LD_LIBRARY_PATH ; "
+                            "DYLD_LIBRARY_PATH=%s%s${DYLD_LIBRARY_PATH:-} ; export DYLD_LIBRARY_PATH ; "
                             "%s %s",
+                            (NULL != mca_plm_rsh_component.chdir ? "cd " : " "),
+                            (NULL != mca_plm_rsh_component.chdir ? mca_plm_rsh_component.chdir : " "),
+                            (NULL != mca_plm_rsh_component.chdir ? " ; " : " "),
                             (opal_prefix != NULL ? "OPAL_PREFIX=" : " "),
                             (opal_prefix != NULL ? opal_prefix : " "),
                             (opal_prefix != NULL ? " ; export OPAL_PREFIX;" : " "),
+                            (opal_libdir != NULL ? "OPAL_LIBDIR=" : " "),
+                            (opal_libdir != NULL ? opal_libdir : " "),
+                            (opal_libdir != NULL ? " ; export OPAL_LIBDIR;" : " "),
                             (NULL != bin_base ? bin_base : " "),
                             (NULL != bin_base ? ":" : " "),
                             (NULL != lib_base ? lib_base : " "),
@@ -530,7 +544,7 @@ static int setup_launch(int *argcptr, char ***argvptr,
              * we have to insert the orted_prefix in the right place
              */
             (void)asprintf (&final_cmd,
-                            "%s%s%s set path = ( %s $path ) ; "
+                            "%s%s%s%s%s%s%s%s%s set path = ( %s $path ) ; "
                             "if ( $?LD_LIBRARY_PATH == 1 ) "
                             "set OMPI_have_llp ; "
                             "if ( $?LD_LIBRARY_PATH == 0 ) "
@@ -544,9 +558,15 @@ static int setup_launch(int *argcptr, char ***argvptr,
                             "if ( $?OMPI_have_dllp == 1 ) "
                             "setenv DYLD_LIBRARY_PATH %s%s$DYLD_LIBRARY_PATH ; "
                             "%s %s",
+                            (NULL != mca_plm_rsh_component.chdir ? "cd " : " "),
+                            (NULL != mca_plm_rsh_component.chdir ? mca_plm_rsh_component.chdir : " "),
+                            (NULL != mca_plm_rsh_component.chdir ? " ; " : " "),
                             (opal_prefix != NULL ? "setenv OPAL_PREFIX " : " "),
                             (opal_prefix != NULL ? opal_prefix : " "),
                             (opal_prefix != NULL ? " ;" : " "),
+                            (opal_libdir != NULL ? "setenv OPAL_LIBDIR " : " "),
+                            (opal_libdir != NULL ? opal_libdir : " "),
+                            (opal_libdir != NULL ? " ;" : " "),
                             (NULL != bin_base ? bin_base : " "),
                             (NULL != lib_base ? lib_base : " "),
                             (NULL != lib_base ? lib_base : " "),
@@ -607,6 +627,19 @@ static int setup_launch(int *argcptr, char ***argvptr,
          (mca_plm_rsh_component.using_qrsh && mca_plm_rsh_component.daemonize_qrsh)) &&
         ((!mca_plm_rsh_component.using_llspawn) ||
          (mca_plm_rsh_component.using_llspawn && mca_plm_rsh_component.daemonize_llspawn))) {
+        opal_argv_append(&argc, &argv, "--daemonize");
+    }
+
+    if (!mca_plm_rsh_component.no_tree_spawn) {
+        // Remove problematic and/or conflicting command line arguments that
+        // should not be passed on to our children.
+        cnt = opal_argv_count(orted_cmd_line);
+        for (i=0; i < cnt; i+=3) {
+            if (0 == strcmp(orted_cmd_line[i+1], "routed")) {
+                opal_argv_delete(&cnt, &orted_cmd_line, i, 3);
+                break;
+            }
+        }
     }
 
     /*
@@ -627,6 +660,16 @@ static int setup_launch(int *argcptr, char ***argvptr,
     if (!mca_plm_rsh_component.no_tree_spawn) {
         opal_argv_append(&argc, &argv, "--tree-spawn");
         orte_oob_base_get_addr(&param);
+
+        // When tree-spawn'ing we need to force the remote daemons to use
+        // the routing component that was used to setup the launch tree.
+        // Otherwise the orte_parent_uri will not match the orted they
+        // expect to find in the routing tree.
+        rtmod = orte_rml.get_routed(orte_coll_conduit);
+        opal_argv_append(&argc, &argv, "-"OPAL_MCA_CMD_LINE_ID);
+        opal_argv_append(&argc, &argv, "routed");
+        opal_argv_append(&argc, &argv, rtmod);
+
         opal_argv_append(&argc, &argv, "-"OPAL_MCA_CMD_LINE_ID);
         opal_argv_append(&argc, &argv, "orte_parent_uri");
         opal_argv_append(&argc, &argv, param);
@@ -1187,6 +1230,10 @@ static void launch_daemons(int fd, short args, void *cbdata)
         OBJ_CONSTRUCT(&coll, opal_list_t);
         rtmod = orte_rml.get_routed(orte_coll_conduit);
         orte_routed.get_routing_list(rtmod, &coll);
+
+        OPAL_OUTPUT_VERBOSE((1, orte_plm_base_framework.framework_output,
+                             "%s plm:rsh:launch Tree Launch using routed/%s",
+                             ORTE_NAME_PRINT(ORTE_PROC_MY_NAME), rtmod));
     }
 
     /* setup the launch */
